from __future__ import annotations

import numpy as np
from spglib import get_symmetry_from_database


def test_change_of_basis(crystal_data_dataset):
    crystal_data = crystal_data_dataset["crystal_data"]
    dataset = crystal_data_dataset["dataset"]
    symprec = crystal_data_dataset["symprec"]
    std_pos = dataset.std_positions
    tmat = dataset.transformation_matrix
    orig_shift = dataset.origin_shift
    lat = np.dot(crystal_data.cell[0].T, np.linalg.inv(tmat))
    pos = np.dot(crystal_data.cell[1], tmat.T) + orig_shift
    for p in pos:
        diff = std_pos - p
        diff -= np.rint(diff)
        diff = np.dot(diff, lat.T)
        delta = np.sqrt((diff**2).sum(axis=1))
        indices = np.where(delta < symprec)[0]
        assert len(indices) == 1


def test_std_symmetry(crystal_data_dataset):
    dataset = crystal_data_dataset["dataset"]
    symmetry = get_symmetry_from_database(dataset.hall_number)
    std_pos = dataset.std_positions

    # for r, t in zip(symmetry['rotations'], symmetry['translations']):
    #     for rp in (np.dot(std_pos, r.T) + t):
    #         diff = std_pos - rp
    #         diff -= np.rint(diff)
    #         num_match = len(np.where(abs(diff).sum(axis=1) < 1e-3)[0])
    #         self.assertEqual(num_match, 1, msg="%s" % fname)

    # Equivalent above by numpy hack
    # 15 sec on macOS 2.3 GHz Intel Core i5 (4times faster than above)
    rot = symmetry["rotations"]
    trans = symmetry["translations"]
    # (n_sym, 3, n_atom)
    rot_pos = np.dot(rot, std_pos.T) + trans[:, :, None]
    for p in std_pos:
        diff = rot_pos - p[None, :, None]
        diff -= np.rint(diff)
        num_match = (abs(diff).sum(axis=1) < 1e-3).sum(axis=1)
        assert all(num_match == 1)


def test_std_rotation(crystal_data_dataset):
    crystal_data = crystal_data_dataset["crystal_data"]
    dataset = crystal_data_dataset["dataset"]
    symprec = crystal_data_dataset["symprec"]
    std_lat = dataset.std_lattice
    tmat = dataset.transformation_matrix
    lat = np.dot(crystal_data.cell[0].T, np.linalg.inv(tmat))
    lat_rot = np.dot(dataset.std_rotation_matrix, lat)
    np.testing.assert_allclose(
        std_lat,
        lat_rot.T,
        atol=symprec,
    )



import importlib.metadata

project = "Spglib"
copyright = "2009, Atsushi Togo"

extensions = [
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.bibtex",
    "myst_parser",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx_tippy",
]

exclude_patterns = [
    "README.md",
    "_build",
]
source_suffix = {
    ".md": "markdown",
    ".rst": "restructuredtext",
}

# -----------------------------------------------------------------------------
# MyST
# -----------------------------------------------------------------------------

myst_enable_extensions = [
    "amsmath",
    "dollarmath",
    "html_admonition",
    "html_image",
    "replacements",
    "smartquotes",
    "tasklist",
    "colon_fence",
]

myst_dmath_double_inline = True
myst_heading_anchors = 3

# -----------------------------------------------------------------------------
# BibTeX
# -----------------------------------------------------------------------------

bibtex_bibfiles = ["references.bib"]
bibtex_default_style = "unsrt"

# -----------------------------------------------------------------------------
# Sphinx-book-theme
# -----------------------------------------------------------------------------

html_theme = "sphinx_book_theme"
html_title = f"Spglib v{importlib.metadata.version('spglib')}"
html_theme_options = {
    # https://sphinx-book-theme.readthedocs.io/en/latest/reference.html
    "repository_url": "https://github.com/spglib/spglib",
    "use_repository_button": True,
    "navigation_with_keys": True,
    "globaltoc_includehidden": "true",
    "show_toc_level": 3,
}
html_static_path = ["_static"]

# -----------------------------------------------------------------------------
# linkcheck
# -----------------------------------------------------------------------------

linkcheck_ignore = [
    # Ignore closed-access links
    r"https://www.sciencedirect.com/.*",
    r"https://www.jstor.org/.*",
    # Cannot check for 403 error only
    r"https://doi.org/10.1002/qua.20747",
    r"https://github.com/spglib/spglib/pull/.*",
    r"https://doi.org/10.1080/27660400.2024.2384822",
]
linkcheck_allowed_redirects = {
    r"https://doi.org/.*": r".*",
    r"https://github.com/spglib/spglib/archive/.*": r".*",
    r"https://spglib.readthedocs.io/.*": r"https://spglib.readthedocs.io/.*",
    r"https://docs.rs/crate/spglib/": r"https://docs.rs/crate/spglib/.*",
    r"https://github.com/chrisjsewell/sphinx-autodoc2": r"https://github.com/sphinx-extensions2/sphinx-autodoc2",
    r"https://github.com/spglib/spglib/tree/.*": r"https://github.com/spglib/spglib/blob/.*",
}
linkcheck_anchors_ignore_for_url = [
    r"https://github.com",
]

# -----------------------------------------------------------------------------
# extlinks
# -----------------------------------------------------------------------------

extlinks = {
    "issue": ("https://github.com/spglib/spglib/issues/%s", "issue %s"),
    "path": ("https://github.com/spglib/spglib/tree/develop/%s", "%s"),
    "user": ("https://github.com/%s", "%s"),
}

# -----------------------------------------------------------------------------
# intersphinx and tippy
# -----------------------------------------------------------------------------

intersphinx_mapping = {
    "cmake": ("https://cmake.org/cmake/help/latest", None),
    "scikit": ("https://scikit-build-core.readthedocs.io/en/latest/", None),
}

tippy_rtd_urls = [
    # Only works with RTD hosted intersphinx
    # "https://cmake.org/cmake/help/latest",
    "https://scikit-build-core.readthedocs.io/en/latest/",
]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from spglib import get_symmetry_dataset

if TYPE_CHECKING:
    from conftest import CrystalData


@pytest.mark.benchmark(group="space-group")
def test_get_symmetry_dataset(
    benchmark,
    crystal_data: CrystalData,
):
    """Benchmarking get_symmetry_dataset on all structures under test/data."""
    # TODO: This benchmarks individual crystal data. Is it useful?
    print(f"Benchmark get_symmetry_dataset on {crystal_data.name} structures")

    def _get_symmetry_dataset_for_cells():
        _ = get_symmetry_dataset(crystal_data.cell, symprec=1e-5)

    benchmark.pedantic(_get_symmetry_dataset_for_cells, rounds=4)



"""Pytest configuration."""

from __future__ import annotations

import dataclasses
import os
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pytest
import spglib
import yaml

if TYPE_CHECKING:
    from typing import Self, TypeAlias

    TestCell: TypeAlias = tuple[np.ndarray, np.ndarray, np.ndarray]


TEST_DIR = Path(__file__).parent


def get_cell(fname: Path) -> TestCell:
    with fname.open() as f:
        yaml_data = yaml.safe_load(f)["unitcell"]
    lattice = np.array(yaml_data["lattice"])
    numbers = np.array([v["number"] for v in yaml_data["points"]], dtype=int)
    points = np.array([v["coordinates"] for v in yaml_data["points"]])
    return lattice, points, numbers


@dataclasses.dataclass()
class CrystalData:
    name: str
    cell: TestCell
    # TODO: Convert this to a TypeDict
    ref: dict | None

    @classmethod
    def from_path(cls, path: str | Path) -> Self:
        if not isinstance(path, Path):
            path = Path(path)
        if path.is_absolute():
            name = path.stem
            cell_file = path
            ref_file = None
        else:
            name = str(path)
            cell_file = TEST_DIR / f"data/{path}.yaml"
            ref_file = TEST_DIR / f"ref/{path}-ref.yaml"
        cell_data = get_cell(cell_file)
        ref_data = None
        if ref_file and ref_file.exists():
            with ref_file.open() as f:
                ref_data = yaml.safe_load(f)
        return cls(
            name=name,
            cell=cell_data,
            ref=ref_data,
        )


@pytest.fixture(scope="session")
def get_crystal_data():
    """Get a single crystal data."""

    def _get_crystal_data(path: str) -> CrystalData:
        return CrystalData.from_path(path)

    return _get_crystal_data


# TODO: Scope here is not correct because it depends on symprec value
@pytest.fixture(scope="session")
def crystal_data_dataset(crystal_data: CrystalData, request: pytest.FixtureRequest):
    other_kwargs = {}
    symprec = request.node.get_closest_marker("symprec") or 1e-5
    angle_tolerance = request.node.get_closest_marker("angle_tolerance")
    if angle_tolerance is not None:
        other_kwargs["angle_tolerance"] = angle_tolerance
    dataset = spglib.get_symmetry_dataset(
        crystal_data.cell,
        **other_kwargs,
    )
    # TODO: reorganize this better
    return {
        "crystal_data": crystal_data,
        "dataset": dataset,
        "symprec": symprec,
    }


def pytest_generate_tests(metafunc: pytest.Metafunc):
    """Generate all crystal data tests tests."""
    if "crystal_data" in metafunc.fixturenames:
        data_root = TEST_DIR / "data"
        all_crystal_data = []
        # Path.walk was introduced in 3.12. For now use os.walk
        for root, dirs, files in os.walk(data_root):
            for file_name in files:
                file_path = Path(root) / file_name
                if file_path.suffix not in (".yaml", ".yml"):
                    continue
                relative_path = file_path.relative_to(data_root)
                # Strip the suffix
                relative_path = relative_path.with_suffix("")
                crystal_data = CrystalData.from_path(relative_path)
                all_crystal_data.append(
                    pytest.param(
                        crystal_data,
                        id=str(relative_path),
                    )
                )
        metafunc.parametrize("crystal_data", all_crystal_data, scope="session")



"""Test of spacegroup_type_from_symmetry."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from spglib import (
    get_spacegroup_type,
    get_spacegroup_type_from_symmetry,
)

if TYPE_CHECKING:
    pass


@pytest.mark.parametrize("lattice_None", [True, False])
def test_spacegroup_type_from_symmetry(
    lattice_None: bool,
    crystal_data_dataset,
):
    """Test spacegroup_type_from_symmetry."""
    if lattice_None:
        lattice = None
    else:
        lattice = crystal_data_dataset["crystal_data"].cell[0]

    dataset = crystal_data_dataset["dataset"]
    spgtype = get_spacegroup_type_from_symmetry(
        dataset.rotations,
        dataset.translations,
        lattice=lattice,
        symprec=crystal_data_dataset["symprec"],
    )

    assert spgtype.number == dataset.number
    spgtype_ref = get_spacegroup_type(dataset.hall_number)
    assert spgtype == spgtype_ref



"""Test of get_hall_number_from_symmetry."""

from __future__ import annotations

import pytest
from spglib import get_hall_number_from_symmetry


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_get_hall_number_from_symmetry(crystal_data_dataset):
    """Test get_hall_number_from_symmetry."""
    dataset = crystal_data_dataset["dataset"]
    hall_number = get_hall_number_from_symmetry(
        dataset.rotations,
        dataset.translations,
        symprec=crystal_data_dataset["symprec"],
    )
    assert hall_number == dataset.hall_number



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

import numpy as np
from spglib import (
    MagneticSpaceGroupType,
    find_primitive,
    get_magnetic_spacegroup_type,
    get_magnetic_symmetry_from_database,
    get_pointgroup,
    get_spacegroup,
    get_spacegroup_type,
    get_symmetry_dataset,
    standardize_cell,
)

if TYPE_CHECKING:
    from conftest import CrystalData

# fmt: off
spg_to_hall = [
    1,   2,   3,   6,   9,   18,  21,  30,  39,  57,  # noqa: E241
    60,  63,  72,  81,  90,  108, 109, 112, 115, 116,  # noqa: E241
    119, 122, 123, 124, 125, 128, 134, 137, 143, 149,
    155, 161, 164, 170, 173, 176, 182, 185, 191, 197,
    203, 209, 212, 215, 218, 221, 227, 228, 230, 233,
    239, 245, 251, 257, 263, 266, 269, 275, 278, 284,
    290, 292, 298, 304, 310, 313, 316, 322, 334, 335,
    337, 338, 341, 343, 349, 350, 351, 352, 353, 354,
    355, 356, 357, 358, 359, 361, 363, 364, 366, 367,
    368, 369, 370, 371, 372, 373, 374, 375, 376, 377,
    378, 379, 380, 381, 382, 383, 384, 385, 386, 387,
    388, 389, 390, 391, 392, 393, 394, 395, 396, 397,
    398, 399, 400, 401, 402, 404, 406, 407, 408, 410,
    412, 413, 414, 416, 418, 419, 420, 422, 424, 425,
    426, 428, 430, 431, 432, 433, 435, 436, 438, 439,
    440, 441, 442, 443, 444, 446, 447, 448, 449, 450,
    452, 454, 455, 456, 457, 458, 460, 462, 463, 464,
    465, 466, 467, 468, 469, 470, 471, 472, 473, 474,
    475, 476, 477, 478, 479, 480, 481, 482, 483, 484,
    485, 486, 487, 488, 489, 490, 491, 492, 493, 494,
    495, 497, 498, 500, 501, 502, 503, 504, 505, 506,
    507, 508, 509, 510, 511, 512, 513, 514, 515, 516,
    517, 518, 520, 521, 523, 524, 525, 527, 529, 530, 531]
# fmt: on


def get_spgnum(name: str) -> int:
    """Get number reference from filename."""
    file_name = pathlib.Path(name).stem
    spgnum = int(file_name.split("_")[1])
    return spgnum


def test_get_symmetry_dataset(crystal_data: CrystalData):
    symprec = 1e-5
    dataset = get_symmetry_dataset(crystal_data.cell, symprec=symprec)
    spgnum = get_spgnum(crystal_data.name)
    assert dataset.number == spgnum

    for i in range(spg_to_hall[spgnum - 1], spg_to_hall[spgnum]):
        dataset_with_hall_number = get_symmetry_dataset(
            crystal_data.cell, hall_number=i, symprec=symprec
        )
        assert dataset_with_hall_number.hall_number == i
        spg_type = get_spacegroup_type(dataset_with_hall_number.hall_number)
        assert dataset_with_hall_number.international == spg_type.international_short
        assert dataset_with_hall_number.hall == spg_type.hall_symbol
        assert dataset_with_hall_number.choice == spg_type.choice
        assert dataset_with_hall_number.pointgroup == spg_type.pointgroup_international

    wyckoffs = dataset.wyckoffs
    assert wyckoffs == crystal_data.ref["wyckoffs"]


def test_standardize_cell_and_pointgroup(crystal_data: CrystalData):
    spgnum = get_spgnum(crystal_data.name)
    symprec = 1e-5
    std_cell = standardize_cell(
        crystal_data.cell,
        to_primitive=False,
        no_idealize=True,
        symprec=symprec,
    )
    dataset = get_symmetry_dataset(std_cell, symprec=symprec)
    assert dataset.number == spgnum

    # The test for point group has to be done after standardization.
    ptg_symbol, _, _ = get_pointgroup(dataset.rotations)
    assert dataset.pointgroup == ptg_symbol


def test_standardize_cell_from_primitive(crystal_data: CrystalData):
    spgnum = get_spgnum(crystal_data.name)
    symprec = 1e-5
    prim_cell = standardize_cell(
        crystal_data.cell,
        to_primitive=True,
        no_idealize=True,
        symprec=symprec,
    )
    std_cell = standardize_cell(
        prim_cell,
        to_primitive=False,
        no_idealize=True,
        symprec=symprec,
    )
    dataset = get_symmetry_dataset(std_cell, symprec=symprec)
    assert dataset.number == spgnum


def test_standardize_cell_to_primitive(crystal_data: CrystalData):
    spgnum = get_spgnum(crystal_data.name)
    symprec = 1e-5
    prim_cell = standardize_cell(
        crystal_data.cell,
        to_primitive=True,
        no_idealize=True,
        symprec=symprec,
    )
    dataset = get_symmetry_dataset(prim_cell, symprec=symprec)
    assert dataset.number == spgnum


def test_refine_cell(crystal_data_dataset):
    crystal_data = crystal_data_dataset["crystal_data"]
    dataset_0 = crystal_data_dataset["dataset"]
    symprec = crystal_data_dataset["symprec"]
    spgnum = get_spgnum(crystal_data.name)
    ref_cell_0 = (
        dataset_0.std_lattice,
        dataset_0.std_positions,
        dataset_0.std_types,
    )
    dataset_1 = get_symmetry_dataset(ref_cell_0, symprec=symprec)
    # Check the same space group type is found.
    assert dataset_1.number == spgnum

    # Check if the same structure is obtained when applying
    # standardization again, i.e., examining non cycling behaviour.
    # Currently only for orthorhombic.

    ref_cell_1 = (
        dataset_1.std_lattice,
        dataset_1.std_positions,
        dataset_1.std_types,
    )
    dataset_2 = get_symmetry_dataset(ref_cell_1, symprec=symprec)
    np.testing.assert_equal(
        dataset_1.std_types,
        dataset_2.std_types,
    )
    np.testing.assert_allclose(
        dataset_1.std_lattice,
        dataset_2.std_lattice,
        atol=symprec,
    )
    diff = dataset_1.std_positions - dataset_2.std_positions
    diff -= np.rint(diff)
    np.testing.assert_allclose(diff, 0, atol=symprec)


def test_get_spacegroup():
    cell = (
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ],
        [
            [0.0, 0.0, 0.0],
            [0.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.5, 0.5, 0.0],
        ],
        [0, 0, 0, 0],
    )
    assert get_spacegroup(cell) == "Fm-3m (225)"
    assert get_spacegroup(cell, symbol_type=1) == "Oh^5 (225)"


def test_find_primitive(crystal_data_dataset):
    crystal_data = crystal_data_dataset["crystal_data"]
    dataset = crystal_data_dataset["dataset"]
    symprec = crystal_data_dataset["symprec"]
    primitive = find_primitive(crystal_data.cell, symprec=symprec)

    spg_type = get_spacegroup_type(dataset.hall_number)
    c = spg_type.international_short[0]
    if c in ["A", "B", "C", "I"]:
        multiplicity = 2
    elif c == "F":
        multiplicity = 4
    elif c == "R":
        assert spg_type.choice == "H"
        if spg_type.choice == "H":
            multiplicity = 3
        else:  # spg_type['choice'] == 'R'
            multiplicity = 1
    else:
        multiplicity = 1
    assert len(dataset.std_types) == len(primitive[2]) * multiplicity


def test_magnetic_spacegroup_type():
    # P 3 -2"
    actual1 = get_magnetic_spacegroup_type(1279)
    expect1 = MagneticSpaceGroupType(
        uni_number=1279,
        litvin_number=1279,
        bns_number="156.49",
        og_number="156.1.1279",
        number=156,
        type=1,
    )
    assert actual1 == expect1

    # -P 2 2ab 1'
    actual2 = get_magnetic_spacegroup_type(452)
    expect2 = MagneticSpaceGroupType(
        uni_number=452,
        litvin_number=442,
        bns_number="55.354",
        og_number="55.2.442",
        number=55,
        type=2,
    )
    assert actual2 == expect2

    # P 31 2 1c' (0 0 1)
    actual3 = get_magnetic_spacegroup_type(1262)
    expect3 = MagneticSpaceGroupType(
        uni_number=1262,
        litvin_number=1270,
        bns_number="151.32",
        og_number="153.4.1270",
        number=151,
        type=4,
    )
    assert actual3 == expect3


def test_magnetic_symmetry_database():
    # UNI: R31'_c[R3] (1242), BNS: R_I3 (146.12)

    # Hexagonal axes: hall_number: 433
    data_h_actual = get_magnetic_symmetry_from_database(1242)
    for key in ["rotations", "translations", "time_reversals"]:
        assert len(data_h_actual[key]) == 18

    # Rhombohedral axes: hall_number: 434
    data_r_actual = get_magnetic_symmetry_from_database(1242, hall_number=434)
    data_r_expect = {
        "rotations": np.array(
            [
                # x,y,z
                [
                    [1, 0, 0],
                    [0, 1, 0],
                    [0, 0, 1],
                ],
                # y,z,x
                [
                    [0, 0, 1],
                    [1, 0, 0],
                    [0, 1, 0],
                ],
                # y+1/2,z+1/2,x+1/2'
                [
                    [0, 0, 1],
                    [1, 0, 0],
                    [0, 1, 0],
                ],
                # z,x,y
                [
                    [0, 1, 0],
                    [0, 0, 1],
                    [1, 0, 0],
                ],
                # x+1/2,y+1/2,z+1/2'
                [
                    [1, 0, 0],
                    [0, 1, 0],
                    [0, 0, 1],
                ],
                # z+1/2,x+1/2,y+1/2'
                [
                    [0, 1, 0],
                    [0, 0, 1],
                    [1, 0, 0],
                ],
            ],
            dtype=np.int32,
        ),
        "translations": np.array(
            [
                [0, 0, 0],
                [0, 0, 0],
                [0.5, 0.5, 0.5],
                [0, 0, 0],
                [0.5, 0.5, 0.5],
                [0.5, 0.5, 0.5],
            ],
        ),
        "time_reversals": np.array(
            [
                [0, 0, 1, 0, 1, 1],
            ],
        ),
    }
    for key in ["rotations", "translations", "time_reversals"]:
        assert np.allclose(data_r_actual[key], data_r_expect[key])



"""Python bindings for C library for finding and handling crystal."""
# Copyright (C) 2015 Atsushi Togo
# This file is part of spglib.
# SPDX-License-Identifier: BSD-3-Clause

from ._version import __version__, __version_tuple__  # noqa: F401

# fmt: off
from .spglib import (  # noqa: F401
    MagneticSpaceGroupType,
    SpaceGroupType,
    SpglibDataset,
    SpglibMagneticDataset,
    delaunay_reduce,
    find_primitive,
    get_BZ_grid_points_by_rotations,
    get_error_message,
    get_grid_point_from_address,
    get_grid_points_by_rotations,
    get_hall_number_from_symmetry,
    get_ir_reciprocal_mesh,
    get_layergroup,
    get_magnetic_spacegroup_type,
    get_magnetic_spacegroup_type_from_symmetry,
    get_magnetic_symmetry,
    get_magnetic_symmetry_dataset,
    get_magnetic_symmetry_from_database,
    get_pointgroup,
    get_spacegroup,
    get_spacegroup_type,
    get_spacegroup_type_from_symmetry,
    get_stabilized_reciprocal_mesh,
    get_symmetry,
    get_symmetry_dataset,
    get_symmetry_from_database,
    get_version,
    niggli_reduce,
    refine_cell,
    relocate_BZ_grid_address,
    spg_get_commit,
    spg_get_version,
    spg_get_version_full,
    standardize_cell,
)

# fmt: on
